@@ -19,12 +19,9 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 
-import gi
 import threading
 import os
 
-gi.require_version('Ide', '1.0')
-
 from gi.repository import Gio
 from gi.repository import GLib
 from gi.repository import GObject
@@ -39,7 +36,14 @@ _ERROR_FORMAT_REGEX = ("(?<filename>[a-zA-Z0-9\\+\\-\\.\\/_]+):"
                        "(?<level>[\\w\\[a-zA-Z0-9\\]\\s]+): "
                        "(?<message>.*)")
 
-class CargoBuildSystem(Ide.Object, Ide.BuildSystem, Gio.AsyncInitable):
+class CargoBuildSystemDiscovery(Ide.SimpleBuildSystemDiscovery):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.props.glob = 'Cargo.toml'
+        self.props.hint = 'cargo_plugin'
+        self.props.priority = -200
+
+class CargoBuildSystem(Ide.Object, Ide.BuildSystem):
     project_file = GObject.Property(type=Gio.File)
 
     def do_get_id(self):
@@ -48,33 +52,6 @@ class CargoBuildSystem(Ide.Object, Ide.BuildSystem, Gio.AsyncInitable):
     def do_get_display_name(self):
         return 'Cargo'
 
-    def do_init_async(self, io_priority, cancellable, callback, data):
-        task = Gio.Task.new(self, cancellable, callback)
-
-        # This is all done synchronously, doing it in a thread would probably
-        # be somewhat ideal although unnecessary at this time.
-
-        try:
-            # Maybe this is a Cargo.toml
-            if self.props.project_file.get_basename() in ('Cargo.toml',):
-                task.return_boolean(True)
-                return
-
-            # Maybe this is a directory with a Cargo.toml
-            if self.props.project_file.query_file_type(0) == Gio.FileType.DIRECTORY:
-                child = self.props.project_file.get_child('Cargo.toml')
-                if child.query_exists(None):
-                    self.props.project_file = child
-                    task.return_boolean(True)
-                    return
-        except Exception as ex:
-            task.return_error(ex)
-
-        task.return_error(Ide.NotSupportedError())
-
-    def do_init_finish(self, task):
-        return task.propagate_boolean()
-
     def do_get_priority(self):
         return -200
 
@@ -92,30 +69,46 @@ def locate_cargo_from_config(config):
 
     return cargo
 
-class CargoPipelineAddin(Ide.Object, Ide.BuildPipelineAddin):
+class CargoPipelineAddin(Ide.Object, Ide.PipelineAddin):
     """
     The CargoPipelineAddin is responsible for creating the necessary build
     stages and attaching them to phases of the build pipeline.
     """
+    error_format_id = None
+    launcher_handler = None
 
-    def do_load(self, pipeline):
-        context = self.get_context()
-        build_system = context.get_build_system()
+    def _on_launcher_created_cb(self, pipeline, launcher):
+        # Set RUSTFLAGS so that we can parse error formats
+        if launcher.getenv('RUSTFLAGS') is None:
+            eq_srcdir = '=' + pipeline.get_srcdir()
+            escaped = GLib.shell_quote(eq_srcdir)
+            flags = '--error-format=short --remap-path-prefix ' + escaped
+            launcher.setenv('RUSTFLAGS', flags, False)
 
-        # Always register the error regex
+    def do_prepare(self, pipeline):
         self.error_format_id = pipeline.add_error_format(_ERROR_FORMAT_REGEX,
                                                          GLib.RegexCompileFlags.OPTIMIZE |
                                                          GLib.RegexCompileFlags.CASELESS);
+        self.launcher_handler = pipeline.connect('launcher-created',
+                                                 self._on_launcher_created_cb)
 
+    def do_load(self, pipeline):
+        context = self.get_context()
+        build_system = Ide.BuildSystem.from_context(context)
 
         # Ignore pipeline unless this is a cargo project
         if type(build_system) != CargoBuildSystem:
             return
 
-        cargo_toml = build_system.props.project_file.get_path()
-        config = pipeline.get_configuration()
+        project_file = build_system.props.project_file
+        if project_file.get_basename() != 'Cargo.toml':
+            project_file = project_file.get_child('Cargo.toml')
+
+        cargo_toml = project_file.get_path()
+        config = pipeline.get_config()
         builddir = pipeline.get_builddir()
         runtime = config.get_runtime()
+        config_opts = config.get_config_opts()
 
         # We might need to use cargo from ~/.cargo/bin
         cargo = locate_cargo_from_config(config)
@@ -127,7 +120,7 @@ class CargoPipelineAddin(Ide.Object, Ide.BuildPipelineAddin):
         fetch_launcher.push_argv('fetch')
         fetch_launcher.push_argv('--manifest-path')
         fetch_launcher.push_argv(cargo_toml)
-        self.track(pipeline.connect_launcher(Ide.BuildPhase.DOWNLOADS, 0, fetch_launcher))
+        self.track(pipeline.attach_launcher(Ide.PipelinePhase.DOWNLOADS, 0, fetch_launcher))
 
         # Now create our launcher to build the project
         build_launcher = pipeline.create_launcher()
@@ -149,11 +142,14 @@ class CargoPipelineAddin(Ide.Object, Ide.BuildPipelineAddin):
         if not config.props.debug:
             build_launcher.push_argv('--release')
 
-        build_launcher.push_argv('--')
-        build_launcher.push_argv('--error-format')
-        build_launcher.push_argv('short')
-        build_launcher.push_argv('--remap-path-prefix')
-        build_launcher.push_argv('=' + pipeline.get_srcdir())
+        # Configure Options get passed to "cargo rustc" because there is no
+        # equivalent "configure stage" for cargo.
+        if config_opts:
+            try:
+                ret, argv = GLib.shell_parse_argv(config_opts)
+                build_launcher.push_args(argv)
+            except Exception as ex:
+                print(repr(ex))
 
         clean_launcher = pipeline.create_launcher()
         clean_launcher.setenv('CARGO_TARGET_DIR', builddir, True)
@@ -162,17 +158,22 @@ class CargoPipelineAddin(Ide.Object, Ide.BuildPipelineAddin):
         clean_launcher.push_argv('--manifest-path')
         clean_launcher.push_argv(cargo_toml)
 
-        build_stage = Ide.BuildStageLauncher.new(context, build_launcher)
+        build_stage = Ide.PipelineStageLauncher.new(context, build_launcher)
         build_stage.set_name(_("Building project"))
         build_stage.set_clean_launcher(clean_launcher)
         build_stage.connect('query', self._query)
-        self.track(pipeline.connect(Ide.BuildPhase.BUILD, 0, build_stage))
+        self.track(pipeline.attach(Ide.PipelinePhase.BUILD, 0, build_stage))
 
     def do_unload(self, pipeline):
-        if self.error_format_id:
+        if self.error_format_id is not None:
             pipeline.remove_error_format(self.error_format_id)
+            self.error_format_id = None
 
-    def _query(self, stage, pipeline, cancellable):
+        if self.launcher_handler is not None:
+            pipeline.disconnect(self.launcher_handler)
+            self.launcher_handler = None
+
+    def _query(self, stage, pipeline, targets, cancellable):
         # Always defer to cargo to check if build is needed
         stage.set_completed(False)
 
@@ -181,6 +182,9 @@ class CargoBuildTarget(Ide.Object, Ide.BuildTarget):
     def do_get_install_directory(self):
         return None
 
+    def do_get_display_name(self):
+        return 'cargo run'
+
     def do_get_name(self):
         return 'cargo-run'
 
@@ -188,18 +192,14 @@ class CargoBuildTarget(Ide.Object, Ide.BuildTarget):
         return 'rust'
 
     def do_get_argv(self):
-        context = self.get_context()
-        config_manager = context.get_configuration_manager()
+        context = self.ref_context()
+        config_manager = Ide.ConfigManager.from_context(context)
         config = config_manager.get_current()
         cargo = locate_cargo_from_config(config)
 
         # Pass the Cargo.toml path so that we don't
         # need to run from the project directory.
-        project_file = context.get_project_file()
-        if project_file.get_basename() == 'Cargo.toml':
-            cargo_toml = project_file.get_path()
-        else:
-            cargo_toml = project_file.get_child('Cargo.toml')
+        cargo_toml = context.ref_workdir().get_child('Cargo.toml').get_path()
 
         return [cargo, 'run', '--manifest-path', cargo_toml]
 
@@ -213,7 +213,7 @@ class CargoBuildTargetProvider(Ide.Object, Ide.BuildTargetProvider):
         task.set_priority(GLib.PRIORITY_LOW)
 
         context = self.get_context()
-        build_system = context.get_build_system()
+        build_system = Ide.BuildSystem.from_context(context)
 
         if type(build_system) != CargoBuildSystem:
             task.return_error(GLib.Error('Not cargo build system',
@@ -221,7 +221,7 @@ class CargoBuildTargetProvider(Ide.Object, Ide.BuildTargetProvider):
                                          code=Gio.IOErrorEnum.NOT_SUPPORTED))
             return
 
-        task.targets = [CargoBuildTarget(context=self.get_context())]
+        task.targets = [CargoBuildTarget()]
         task.return_boolean(True)
 
     def do_get_targets_finish(self, result):
@@ -235,14 +235,14 @@ class CargoDependencyUpdater(Ide.Object, Ide.DependencyUpdater):
         task.set_priority(GLib.PRIORITY_LOW)
 
         context = self.get_context()
-        build_system = context.get_build_system()
+        build_system = Ide.BuildSystem.from_context(context)
 
         # Short circuit if not using cargo
         if type(build_system) != CargoBuildSystem:
             task.return_boolean(True)
             return
 
-        build_manager = context.get_build_manager()
+        build_manager = Ide.BuildManager.from_context(context)
         pipeline = build_manager.get_pipeline()
         if not pipeline:
             task.return_error(GLib.Error('Cannot update dependencies without build pipeline',
@@ -250,11 +250,14 @@ class CargoDependencyUpdater(Ide.Object, Ide.DependencyUpdater):
                                          code=Gio.IOErrorEnum.FAILED))
             return
 
-        config_manager = context.get_configuration_manager()
+        config_manager = Ide.ConfigManager.from_context(context)
         config = config_manager.get_current()
         cargo = locate_cargo_from_config(config)
 
-        cargo_toml = build_system.props.project_file.get_path()
+        project_file = build_system.props.project_file
+        if project_file.get_basename() != 'Cargo.toml':
+            project_file = project_file.get_child('Cargo.toml')
+        cargo_toml = project_file.get_path()
 
         launcher = pipeline.create_launcher()
         launcher.setenv('CARGO_TARGET_DIR', pipeline.get_builddir(), True)
